import uuid
from http import HTTPStatus

from PyQt5.QtCore import QVariant
from osgeo import ogr, gdal
from qgis.PyQt.QtWidgets import QDialog
from qgis._core import QgsLineString
from qgis.core import (
    QgsApplication,
    QgsCoordinateReferenceSystem,
    QgsCoordinateTransform,
    QgsFeature,
    QgsField,
    QgsProject,
    QgsVectorLayer,
)
from qgis.utils import iface

from .panels import Panels
from ..tools.tools import MyRectangleTool
from ..tools.networking import NetworkRequestTask


class TopologyExtractPanel:
    def __init__(self, dialog: QDialog, plugin) -> None:
        self._plugin = plugin
        self._dialog = dialog

        # self.oldtool = None
        self.startPoint = self.endPoint = None
        self.isDownloading = False

    def setup_panel(self) -> None:
        print("Setup topology extract panel")

        self.display_error()

        self._dialog.topology_extract_click_button.clicked.connect(
            self.clicked_btn_rectangle_tool
        )
        self._dialog.topology_extract_download_button.clicked.connect(
            self.clicked_btn_download
        )

    def make_active_panel(self):
        print("isochrone make active")
        self.display_error()

    def display_error(self):
        if self._plugin.error_message is None:
            self._dialog.topology_extract_error_label.setVisible(False)
        else:
            self._dialog.topology_extract_error_label.setText(
                self._plugin.error_message
            )
            self._dialog.topology_extract_error_label.setVisible(True)

    def clicked_btn_rectangle_tool(self):
        canvas = iface.mapCanvas()
        # self.oldtool = canvas.mapTool()
        my_rectangle_tool = MyRectangleTool(
            canvas, callback_rectangle_created=self.callback_rectangle_created
        )
        canvas.setMapTool(my_rectangle_tool)

    def callback_rectangle_created(self, my_rectangle_tool: MyRectangleTool):
        sourceCrs = QgsProject.instance().crs()
        destCrs = QgsCoordinateReferenceSystem.fromEpsgId(4326)
        tr = QgsCoordinateTransform(sourceCrs, destCrs, QgsProject.instance())

        self.startPoint = tr.transform(my_rectangle_tool.startPoint)
        self.endPoint = tr.transform(my_rectangle_tool.endPoint)

        self._dialog.topology_extract_start_x.setText(f"{self.startPoint.x():.6f}")
        self._dialog.topology_extract_start_y.setText(f"{self.startPoint.y():.6f}")
        self._dialog.topology_extract_end_x.setText(f"{self.endPoint.x():.6f}")
        self._dialog.topology_extract_end_y.setText(f"{self.endPoint.y():.6f}")

    def clicked_btn_download(self):
        if self.isDownloading:
            print("wait for download to complete then press again")
            return

        if self.startPoint is None or self.endPoint is None:
            print("Select an area first")
            return

        print("clicked_btn_download")

        point1 = [self.startPoint.x(), self.startPoint.y()]
        point2 = [self.startPoint.x(), self.endPoint.y()]
        point3 = [self.endPoint.x(), self.endPoint.y()]
        point4 = [self.endPoint.x(), self.startPoint.y()]

        url = self._plugin.api_url + "/routing/v1/topology_extract.gpkg"
        request_body = {
            "geometry": {
                "type": "Polygon",
                "coordinates": [[point1, point2, point3, point4, point1]],
                "crs": {
                    "type": "name",
                    "properties": {"name": "urn:ogc:def:crs:EPSG::4326"},
                },
            }
        }
        headers = {
            "Content-type": "application/json",
            "Authorization": self._plugin.api_key,
        }

        self.isDownloading = True
        self._dialog.topology_extract_download_button.setText(
            "Downloading... Please wait"
        )

        globals()["stratopo_topology_extract_download_task"] = NetworkRequestTask(
            url,
            request_headers=headers,
            request_body=request_body,
            callback=self.callback_download_task_finished,
        )
        QgsApplication.taskManager().addTask(
            globals()["stratopo_topology_extract_download_task"]
        )

    def callback_download_task_finished(self, response):
        # TODO extract geopackage from response.data and import layers into project

        if response.status_code == HTTPStatus.OK:
            tmp_file = f"/vsimem/{uuid.uuid4()}.gpkg"
            gdal.FileFromMemBuffer(tmp_file, response.content)
            conn = ogr.Open(tmp_file)

            layers_in_gpkg = {
                "topologies.bikenet_latest": "bikenet_latest",
                "topologies.carnet_latest": "carnet_latest",
                "topologies.walknet_latest": "walknet_latest",
            }

            layers = []

            crs = None

            for gpkg_layer_name, qgis_layer_name in layers_in_gpkg.items():
                layer = QgsVectorLayer(
                    "LineString?crs=epsg:28992", qgis_layer_name, "memory"
                )

                provider = layer.dataProvider()
                provider.addAttributes(
                    [
                        QgsField("gid", QVariant.String),
                        QgsField("osm_id", QVariant.String),
                        QgsField("length_m", QVariant.String),
                        QgsField("source", QVariant.String),
                        QgsField("target", QVariant.String),
                        QgsField("speed", QVariant.String),
                        QgsField("cost", QVariant.String),
                        QgsField("reverse_speed", QVariant.String),
                        QgsField("reverse_cost", QVariant.String),
                    ]
                )
                layer.updateFields()

                gpkg_layer = conn.GetLayer(gpkg_layer_name)
                for f in gpkg_layer:
                    properties = f.items()
                    geometry_wkt = f.geometry().ExportToWkt()

                    feature = QgsFeature(layer.fields())
                    linestring = QgsLineString()
                    linestring.fromWkt(geometry_wkt)
                    feature.setGeometry(linestring)
                    for field in provider.fields():
                        feature.setAttribute(
                            field.name(), properties.get(field.name(), None)
                        )

                    provider.addFeature(feature)

                    # TODO get crs from geopackage (not from geometry)
                    if not crs:
                        try:
                            srid = (
                                f.geometry()
                                .GetSpatialReference()
                                .GetAuthorityCode(None)
                            )
                            crs = QgsCoordinateReferenceSystem.fromEpsgId(int(srid))
                        except ValueError:
                            pass

                layer.updateExtents()

                layers.append(layer)

            if crs:
                for layer in layers:
                    layer.setCrs(crs, emitSignal=False)
                    layer.updateExtents()

            QgsProject.instance().addMapLayers(layers)

            self._plugin.error_message = None
            self.display_error()

        elif response.status_code in (HTTPStatus.FORBIDDEN, HTTPStatus.UNAUTHORIZED):
            print("Error form server: ", response.status_code)
            self._plugin.error_message = "Provide a valid API key and try again"
            self._dialog.switch_panel(Panels.Settings)
        elif response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
            self._plugin.error_message = "Selected area cannot be bigger than 50 km2"
            self.display_error()
        else:
            print("Error form server: ", response.status_code)
            self._plugin.error_message = "Server error"
            self.display_error()

        self.isDownloading = False
        self._dialog.topology_extract_download_button.setText(
            "Download topology within rectangle"
        )
